/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef DATAFRAME_DATAFRAME_HPP
#define DATAFRAME_DATAFRAME_HPP 1

#include <dataframe/dataframe_fwd.hpp>
#include <iosfwd>
#include <map>
#include <stdexcept>
#include <string>
#include <tuple>
#include <typeindex>
#include <typeinfo>
#include <vector>
#include <boost/any.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/throw_exception.hpp>
#include <dataframe/dataframe_reader_fwd.hpp>
#include <dataframe/factor.hpp>
#include <dataframe/row_iterator.hpp>
#include <dataframe/sequence_proxy.hpp>

namespace dataframe
{

  /// A string-based dataframe
  /// \tparam Policy The policy for missing or unexpected fields
  ///
  /// \todo Improve error checking
  /// \todo Optionally strip delimiting quotations
  /// \todo Separate dataframe IO from the data structure
  ///
  template < >
  class dataframe< void >
  {
    template < typename Policy >
    friend class dataframe_reader;

    using self = dataframe;

    using string_type = std::string;
    using vector_type = std::vector<string_type>;
    using vector_vector_type = std::vector<vector_type>;

    using element_type = string_type;

    using field_type = vector_type;
    using data_type = vector_vector_type;
    
    using headers_type = vector_type;

    using line_type = string_type;
    using parsed_line_type = vector_type;

    using row_type = vector_type;

  public:
    using size_type = typename vector_type::size_type;
    using header_type = string_type;

  private:
    using key_type = std::pair<size_type,std::type_index>;
    using factor_map_type = std::map<key_type,boost::any>;

  private:
    struct header_not_found
      : std::logic_error
      , virtual boost::exception
    {
      header_not_found ()
	: std::logic_error("header not found")
      {}
      header_not_found (header_type header)
	: std::logic_error(("header '" + header + "' not found").c_str())
      {}
    };

    struct factor_not_found
      : std::logic_error
      , virtual boost::exception
    {
      factor_not_found ()
	: std::logic_error("factor not found; did you forget to create a factor with dataframe.factor<T>(field)?")
      {}
    };

  public:
    bool empty () const { return data_.empty() || data_[0].empty(); }
    size_type size () const { return data_.empty() ? 0 : data_[0].size(); }
    size_type fields () const { return headers_.size(); }
    
  public:
    auto headers () const { return make_sequence_proxy<header_type>(headers_); }
    auto& header (size_type j) { return headers_.at(j); }
    auto& header (size_type j) const { return headers_.at(j); }

  public:
    auto field (size_type j) const
    { return make_sequence_proxy<element_type>(data_.at(j)); }
    
    auto field (header_type header) const
    { return field(find_header(header)); }

    template < typename T >
    auto field (size_type j) const
    { return make_sequence_proxy<T>(data_.at(j)); }

    template < typename T >
    auto field (header_type header) const
    { return field<T>(find_header(header)); }

  private:
    auto make_element (size_type i, size_type j) const
    { return field(j).at(i); }

    template < typename T >
    auto make_element (size_type i, size_type j) const
    { return make_element(T(),i,j); }

    template < typename T >
    auto make_element (T, size_type i, size_type j) const
    { return field(j).template at<T>(i); }

    template < typename T >
    auto make_element (factor<T>, size_type i, size_type j) const
    { return factor<T>(i,j); }

  public:
    auto element (size_type i, size_type j) const
    { return make_element(i,j); }

    auto element (size_type i, header_type header) const
    { return element(i,find_header(header)); }

    template < typename T >
    auto element (size_type i, size_type j) const
    { return make_element<T>(i,j); }

    template < typename T >
    auto element (size_type i, header_type header) const
    { return element<T>(i,find_header(header)); }

  public:
    auto operator () (size_type j) const { return field(j); }
    auto operator () (header_type header) const { return field(header); }

    auto operator () (size_type i, size_type j) const { return element(i,j); }
    auto operator () (size_type i, header_type header) const { return element(i,header); }

  public:
    template < typename ... Args >
    self& insert (header_type header, Args&& ... args)
    {
      headers_.push_back(header);
      data_.push_back(field_type(std::forward<Args>(args)...));
      remove_all_factors();
      return *this;
    }

    self& insert (header_type header, std::initializer_list<element_type> list)
    {
      headers_.push_back(header);
      data_.push_back(field_type(list));
      remove_all_factors();
      return *this;
    }

    template < typename T >
    self& insert (header_type header, std::initializer_list<T> list)
    {
      using boost::lexical_cast;
      field_type field;
      for (auto x : list)
	field.push_back(lexical_cast<element_type>(x));
      headers_.push_back(header);
      data_.push_back(field);
      remove_all_factors();
      return *this;
    }

    self& insert_row (row_type row)
    {
      assert(row.size() == data_.size());
      for (size_type i = 0; i < row.size(); ++i)
	data_[i].push_back(row[i]);
      remove_all_factors();
      return *this;
    }

    self& erase (header_type header)
    {
      auto offset = find_header(header);
      remove_all_factors();
      headers_.erase(headers_.begin() + offset);
      data_.erase(data_.begin() + offset);
      return *this;
    }

    self& erase_row (size_type row)
    {
      assert(row < size());
      remove_all_factors();
      for (auto& field : data_)
	field.erase(field.begin() + row);
      return *this;
    }

  public:
    template < typename T >
    auto& factor (size_type j)
    {
      using factor_type = ::dataframe::factor<T>;
      auto key = std::make_pair(j,std::type_index(typeid(T)));
      auto i = factors_.find(key);
      if (i == factors_.end())
	{
	  auto factor = make_factor<T>(field(j));
	  static_assert(std::is_same<decltype(factor),factor_type>::value,
			"ERROR: mismatched factor types");
	  bool inserted;
	  std::tie(i,inserted) = factors_.insert(std::make_pair(key,factor));
	  assert(inserted);
	}
      return boost::any_cast<factor_type&>(i->second);
    }

    template < typename T >
    auto& factor (header_type header)
    { return factor<T>(find_header(header)); }

    template < typename T >
    auto& factor (size_type j) const
    {
      using factor_type = ::dataframe::factor<T>;
      auto key = std::make_pair(j,std::type_index(typeid(T)));
      auto i = factors_.find(key);
      if (i == factors_.end())
	BOOST_THROW_EXCEPTION(factor_not_found());
      auto& factor = boost::any_cast<factor_type const&>(i->second);
      return factor;
    }

    template < typename T >
    auto& factor (header_type header) const
    { return factor<T>(find_header(header)); }

    template < typename T >
    auto factor (size_type i, size_type j)
    {
      auto factor = this->factor<T>(j);
      auto value = field(j).template at<T>(i).value();
      factor.value(value);
      return factor;
    }

    template < typename T >
    auto factor (size_type i, header_type header)
    { return factor<T>(i,find_header(header)); }

    template < typename T >
    auto factor (size_type i, size_type j) const
    {
      auto factor = this->factor<T>(j);
      auto value = field(j).template at<T>(i).value();
      factor.value(value);
      return factor;
    }

    template < typename T >
    auto factor (size_type i, header_type header) const
    { return factor<T>(i,find_header(header)); }

  public:
    auto& factor (size_type j) { return factor<element_type>(j); }
    auto& factor (header_type header) { return factor<element_type>(header); }
    auto& factor (size_type j) const { return factor<element_type>(j); }
    auto& factor (header_type header) const { return factor<element_type>(header); }

    auto factor (size_type i, size_type j) { return factor<element_type>(i,j); }
    auto factor (size_type i, header_type header) { return factor<element_type>(i,header); }
    auto factor (size_type i, size_type j) const { return factor<element_type>(i,j); }
    auto factor (size_type i, header_type header) const { return factor<element_type>(i,header); }

  private:
    size_type find_header (header_type header) const
    {
      size_type j = 0;
      for (auto hi = headers_.begin(); hi != headers_.end(); ++hi, ++j)
	{
	  if (*hi == header)
	    return j;
	}
      BOOST_THROW_EXCEPTION(header_not_found(header));
    }

    void remove_all_factors () { factors_.clear(); }

  private:
    headers_type headers_;
    data_type data_;
    factor_map_type factors_;
  };

  template < typename Classes >
  auto begin (dataframe<Classes> const& dataframe)
  { return make_row_iterator(dataframe); }

  template < typename Classes >
  auto end (dataframe<Classes> const&)
  { return row_iterator<dataframe<Classes>>(); }

  auto make_dataframe ()
  {
    using dataframe_type = dataframe<>;
    return dataframe_type();
  }

} // namespace dataframe

#endif
