/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_DISTRIBUTION_MCMC_METROPOLIS_SAMPLER_STATE_HPP
#define GRAPH_MODEL_DISTRIBUTION_MCMC_METROPOLIS_SAMPLER_STATE_HPP 1

#include <iosfwd>

namespace graph_model
{

  namespace MCMC
  {

    template < typename Probability, typename Real >
    class metropolis_sampler_state
    {
      using self = metropolis_sampler_state;

    public:
      metropolis_sampler_state () = default;
      metropolis_sampler_state (Probability current_pdf, Real r)
	: current_pdf_(current_pdf)
	, r_(r)
      {}

      auto current_pdf () const { return current_pdf_; }
      auto proposal_pdf () const { return proposal_pdf_; }
      auto alpha () const { return proposal_pdf_ / current_pdf_; }
      auto r () const { return r_; }
      auto accept () const { return r_ < alpha(); }
      auto reject () const { return !accept(); }

      self& current_pdf (Probability pdf) { current_pdf_ = pdf; return *this; }
      self& proposal_pdf (Probability pdf) { proposal_pdf_ = pdf; return *this; }
      self& r (Real r) { r_ = r; return *this; }

      friend std::ostream& operator << (std::ostream& os, self s)
      {
	return os
	  << "metropolis_sampler_state("
	  << "Pr(current)=" << s.current_pdf_ << ","
	  << "Pr(proposal)=" << s.proposal_pdf_ << ","
	  << "alpha=" << s.alpha() << ","
	  << "r=" << s.r_ << ","
	  << "accept=" << std::boolalpha << s.accept()
	  << ")";
      }

    private:
      Probability current_pdf_;
      Probability proposal_pdf_;
      Real r_;
    };

    template < typename Probability, typename Real >
    auto make_metropolis_sampler_state (Probability current_pdf, Real r)
    {
      using state_type = metropolis_sampler_state<Probability,Real>;
      return state_type(current_pdf,r);
    }

  } // namespace MCMC

} // namespace graph_model

#endif
