/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_DISTRIBUTION_UNIFORM_UNIFORM_STRATEGY_HPP
#define GRAPH_MODEL_DISTRIBUTION_UNIFORM_UNIFORM_STRATEGY_HPP 1

#include <type_traits>
#include <boost/math/distributions/uniform.hpp>
#include <boost/random/uniform_real.hpp>
#include <graph_model/distribution/distribution_policy.hpp>
#include <graph_model/distribution/parameter/lower_bound.hpp>
#include <graph_model/distribution/parameter/parameter_pack.hpp>
#include <graph_model/distribution/parameter/upper_bound.hpp>
#include <graph_model/distribution/parameter/variate.hpp>
#include <graph_model/random/distribution.hpp>
#include <graph_model/traits/real.hpp>
#include <sample_space/is_possible.hpp>

namespace graph_model
{

  namespace distribution
  {

    namespace uniform
    {

      /// \brief Generate a uniform distribution
      ///
      /// This policy class models the Distribution Generator concept,
      /// which generates a distribution from its parameters.  A
      /// distribution models the corresponding concept in the Boost
      /// Math library.  Specifically, it can be evaluated by the \c
      /// pdf() function (and possibly the \c cdf() function if it is
      /// a scalar distribution).
      ///
      /// This policy generates a uniform distribution from the Boost
      /// Math library.
      ///
      struct distribution_policy
      {
	template < typename Variate, typename LowerBound, typename UpperBound >
	static auto apply (LowerBound lower_bound, UpperBound upper_bound)
	{
	  using real_type
	    = std::common_type_t<
	      traits::real_t<Variate>
	    , traits::real_t<LowerBound>
	    , traits::real_t<UpperBound>
	    >;
	  using uniform_distribution
	    = boost::math::uniform_distribution<real_type>;
	  return uniform_distribution(lower_bound,upper_bound);
	}
      };

      /// \brief Generate a uniform random number generator
      ///
      /// This policy class models the RNG Generator concept, which
      /// generates a random number distribution from its parameters.
      /// A random number distribution models the corresponding
      /// concept in the Boost Random library.  Specifically, it can
      /// be evaluated to obtain a sample from the distribution.
      ///
      /// This policy generates a uniform distribution from the Boost
      /// Random library.
      ///
      struct generator_policy
      {
	template < typename LowerBound, typename UpperBound >
	static auto& apply ()
	{
	  using real_type
	    = std::common_type_t<
	      traits::real_t<LowerBound>
	    , traits::real_t<UpperBound>
	    >;
	  using uniform_distribution
	    = random::distribution< boost::uniform_real<real_type> >;
	  return uniform_distribution::apply();
	}
      };

      /// \brief Strategy for uniform distributions
      /// \tparam ParameterPack The parameter pack containing the parameters
      ///
      /// A uniform distribution is a distribution across a variate
      /// that is constructible from a real value.  Generally, the
      /// implementation simply delegates to the appropriate
      /// components of the Boost Math or Random libraries.
      ///
      template < typename ParameterPack >
      class uniform_strategy
      {
	using policy_type = distribution::distribution_policy<distribution_policy,generator_policy>;

      public:
	uniform_strategy () = default;
	template < typename Tags, typename Defaults, typename Tagged, typename Untagged >
	explicit uniform_strategy (parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack)
	  : pack_(pack)
	{}

      public:
	template < typename ... T >
	auto variate (T&& ... t) const
	{ return pack_.template value<parameter::tag::variate>(std::forward<T>(t)...); }

	template < typename ... T >
	auto lower_bound (T&& ... t) const
	{ return pack_.template value<parameter::tag::lower_bound>(std::forward<T>(t)...); }

	template < typename ... T >
	auto upper_bound (T&& ... t) const
	{ return pack_.template value<parameter::tag::upper_bound>(std::forward<T>(t)...); }

      public:
	template < typename ... T >
	auto sample (T&& ... t) const
	{
	  return sample(lower_bound(std::forward<T>(t)...),upper_bound(std::forward<T>(t)...));
	}

	template < typename ... T >
	auto pdf (T&& ... t) const
	{
	  return pdf
	    (variate(std::forward<T>(t)...),
	     lower_bound(std::forward<T>(t)...),
	     upper_bound(std::forward<T>(t)...));
	}

      private:
	template < typename LowerBound, typename UpperBound >
	auto sample (LowerBound lower_bound, UpperBound upper_bound) const
	{
	  using sample_space::is_possible;
	  assert(is_possible(lower_bound));
	  assert(is_possible(upper_bound));
	  return policy_type::sample(lower_bound,upper_bound);
	}

	template < typename Variate, typename LowerBound, typename UpperBound >
	auto pdf (Variate variate, LowerBound lower_bound, UpperBound upper_bound) const
	{
	  using sample_space::is_possible;
	  return
	    is_possible(lower_bound) && is_possible(upper_bound)
	    ? policy_type::eval_pdf(variate,lower_bound,upper_bound)
	    : 0;
	}

      private:
	ParameterPack pack_;
      };

      template < typename ParameterPack >
      auto make_uniform_strategy (ParameterPack parameter_pack)
      {
	using uniform_strategy_type = uniform_strategy<ParameterPack>;
	return uniform_strategy_type(parameter_pack);
      };

    } // namespace uniform

  } // namespace distribution

} // namespace graph_model

#endif
