/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_GRAPH_SHARED_NEIGHBOR_GRAPH_HPP
#define GRAPH_MODEL_GRAPH_SHARED_NEIGHBOR_GRAPH_HPP 1

#include <map>
#include <vector>
#include <boost/graph/graph_traits.hpp>
#include <graph_model/graph/vertex_category_property_map_fwd.hpp>
#include <graph_model/graph/vertex_gid_property_map_fwd.hpp>
#include <graph_model/graph/vertex_identifier_property_map_fwd.hpp>
#include <graph_model/graph/vertex_predicates.hpp>
#include <graph_model/graph/undirected_graph.hpp>

namespace graph_model
{

  namespace graph
  {

    namespace tag
    {

      class shared_neighbor {};

    } // namespace tag

    template < typename Graph >
    class vertex_descriptor_property
    {
      using graph_traits = boost::graph_traits<Graph>;
      using vertex_descriptor_type = typename graph_traits::vertex_descriptor;

    public:
      auto& operator = (vertex_descriptor_type v) { vertex_descriptor_ = v; return *this; }

      auto vertex_descriptor () const { return vertex_descriptor_; }

    private:
      vertex_descriptor_type vertex_descriptor_;
    };

    template < typename Graph >
    struct shared_neighbor_graph
      : undirected_graph< tag::shared_neighbor,vertex_descriptor_property<Graph> >
    {};

    template < typename Graph >
    class vertex_category_property_map< shared_neighbor_graph<Graph> >
    {
      using self = vertex_category_property_map;
      using graph_type = shared_neighbor_graph<Graph>;
      using graph_traits = boost::graph_traits<graph_type>;
      using vertex_descriptor_type = typename graph_traits::vertex_descriptor;
      using pointer = observer_ptr<graph_type const>;

    public:
      vertex_category_property_map () = default;
      vertex_category_property_map (graph_type const& g)
	: p_(make_observer(g))
      {}

      friend auto get (self const& map, vertex_descriptor_type u) { return map.vertex_category(u); }

    public:
      using category = boost::readable_property_map_tag;
      using key_type = vertex_descriptor_type;
      using value_type = decltype(std::declval<graph_type>()[std::declval<key_type>()].vertex_descriptor().vertex_category());
      using reference = value_type;

    private:
      auto& graph () { return *p_; }
      auto& graph () const { return *p_; }

      auto vertex_descriptor (vertex_descriptor_type u) const { return graph()[u].vertex_descriptor(); }
      auto vertex_category (vertex_descriptor_type u) const { return vertex_descriptor(u).vertex_category(); }

    private:
      pointer p_;
    };

    template < typename Graph >
    auto get (tag::vertex_category, shared_neighbor_graph<Graph> const& g)
    {
      using graph_type = shared_neighbor_graph<Graph>;
      using property_map = vertex_category_property_map<graph_type>;
      return property_map(g);
    }

    template < typename Graph, typename VertexDescriptor >
    auto get (tag::vertex_category property_tag, shared_neighbor_graph<Graph> const& g, VertexDescriptor u)
    { return get(get(property_tag,g),u); }

    template < typename Graph >
    class vertex_gid_property_map< shared_neighbor_graph<Graph> >
    {
      using self = vertex_gid_property_map;
      using graph_type = shared_neighbor_graph<Graph>;
      using graph_traits = boost::graph_traits<graph_type>;
      using vertex_descriptor_type = typename graph_traits::vertex_descriptor;
      using pointer = observer_ptr<graph_type const>;

    public:
      vertex_gid_property_map () = default;
      vertex_gid_property_map (graph_type const& g)
	: p_(make_observer(g))
      {}

      friend auto get (self const& map, vertex_descriptor_type u)
      { return map.gid(u); }

    public:
      using category = boost::readable_property_map_tag;
      using key_type = vertex_descriptor_type;
      using value_type = decltype(std::declval<graph_type>()[std::declval<key_type>()].vertex_descriptor().id());
      using reference = value_type;

    private:
      auto& graph () { return *p_; }
      auto& graph () const { return *p_; }

      auto vertex_descriptor (vertex_descriptor_type u) const { return graph()[u].vertex_descriptor(); }
      auto gid (vertex_descriptor_type u) const { return vertex_descriptor(u).id(); }

    private:
      pointer p_;
    };

    template < typename Graph >
    auto get (tag::vertex_gid, shared_neighbor_graph<Graph> const& g)
    {
      using graph_type = shared_neighbor_graph<Graph>;
      using property_map = vertex_gid_property_map<graph_type>;
      return property_map(g);
    }

    template < typename Graph, typename VertexDescriptor >
    auto get (tag::vertex_gid property_tag, shared_neighbor_graph<Graph> const& g, VertexDescriptor u)
    { return get(get(property_tag,g),u); }

    template < typename Graph >
    class vertex_identifier_property_map< shared_neighbor_graph<Graph> >
    {
      using self = vertex_identifier_property_map;
      using graph_type = shared_neighbor_graph<Graph>;
      using graph_traits = boost::graph_traits<graph_type>;
      using vertex_descriptor_type = typename graph_traits::vertex_descriptor;
      using pointer = observer_ptr<graph_type const>;

    public:
      vertex_identifier_property_map () = default;
      vertex_identifier_property_map (graph_type const& g)
	: p_(make_observer(g))
      {}

      friend auto get (self const& map, vertex_descriptor_type u) { return map.identifier(u); }

    public:
      using category = boost::readable_property_map_tag;
      using key_type = vertex_descriptor_type;
      using value_type = decltype(std::declval<graph_type>()[std::declval<key_type>()].vertex_descriptor().identifier());
      using reference = value_type;

    private:
      auto& graph () { return *p_; }
      auto& graph () const { return *p_; }

      auto vertex_descriptor (vertex_descriptor_type u) const { return graph()[u].vertex_descriptor(); }
      auto identifier (vertex_descriptor_type u) const { return vertex_descriptor(u).identifier(); }

    private:
      pointer p_;
    };

    template < typename Graph >
    auto get (tag::vertex_identifier, shared_neighbor_graph<Graph> const& g)
    {
      using graph_type = shared_neighbor_graph<Graph>;
      using property_map = vertex_identifier_property_map<graph_type>;
      return property_map(g);
    }

    template < typename Graph, typename VertexDescriptor >
    auto get (tag::vertex_identifier property_tag, shared_neighbor_graph<Graph> const& g, VertexDescriptor u)
    { return get(get(property_tag,g),u); }

    template < typename Graph, typename VertexPredicate, typename NeighborPredicate  >
    auto make_shared_neighbor_graph (Graph const& graph, VertexPredicate vertex_predicate, NeighborPredicate neighbor_predicate)
    {
      using graph_traits = boost::graph_traits<Graph>;
      using vertex_descriptor = typename graph_traits::vertex_descriptor;
      using vector_type = std::vector<vertex_descriptor>;
      using map_type = std::map<vertex_descriptor,vector_type>;

      using vertex_graph_type = boost::filtered_graph<Graph,always_true,VertexPredicate>;
      using shared_neighbor_graph_type = shared_neighbor_graph<Graph>;

      using boost::empty;

      shared_neighbor_graph_type shared_neighbor_graph;
      map_type map;
      vertex_graph_type vertex_graph(graph,always_true(),vertex_predicate);
      for (auto vertex_range = vertices(vertex_graph);
	   !empty(vertex_range);
	   ++vertex_range.first)
	{
	  auto u = *vertex_range.first;
	  auto u2 = add_vertex(shared_neighbor_graph);
	  shared_neighbor_graph[u2] = u;
	  for (auto adjacency_range = adjacent_vertices(u,graph);
	       !empty(adjacency_range);
	       ++adjacency_range.first)
	    {
	      auto v = *adjacency_range.first;
	      assert(neighbor_predicate(v));
	      auto i = map.find(v);
	      if (i != map.end())
		{
		  auto& vector = i->second;
		  std::for_each(vector.begin(),vector.end(),
				[&shared_neighbor_graph,u2,v](auto w)
				{
				  auto w2 = add_vertex(shared_neighbor_graph);
				  shared_neighbor_graph[w2] = w;
				  add_edge(u2,w2,shared_neighbor_graph);
				});
		}
	      map[v].push_back(u);
	    }
	}
      return shared_neighbor_graph;
    }

    template < typename VertexPredicate, typename Graph >
    auto make_shared_neighbor_graph (Graph const& graph)
    { return make_shared_neighbor_graph(graph,VertexPredicate(),always_true()); }

    template < typename VertexPredicate, typename NeighborPredicate, typename Graph >
    auto make_shared_neighbor_graph (Graph const& graph)
    { return make_shared_neighbor_graph(graph,VertexPredicate(),NeighborPredicate()); }

  } // namespace graph

} // namespace graph_model

namespace boost
{

  template < typename Graph >
  struct graph_traits< graph_model::graph::shared_neighbor_graph<Graph> >
    : boost::graph_traits<
        graph_model::graph::undirected_graph<
          graph_model::graph::tag::shared_neighbor
	  , graph_model::graph::vertex_descriptor_property<Graph>
	  >
    >
  {};

} // namespace boost

#endif
