/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#define BOOST_TEST_MODULE dataframe_dataframe

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;

#include "dataframe/dataframe.hpp"
#include "dataframe/row_proxy.hpp"
#include "dataframe/read_dataframe.hpp"
#include "dataframe/write_dataframe.hpp"
#include "algorithm/unique.hpp"
#include <string>
#include <vector>
#include <boost/filesystem/fstream.hpp>
#include <boost/mpl/vector.hpp>
#include <boost/range/algorithm/equal.hpp>

using test_types = boost::mpl::vector<float,double,long double>;

BOOST_AUTO_TEST_CASE_TEMPLATE(test_typedef,T,test_types)
{
  using dataframe_type = dataframe::dataframe<>;
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_default_constructor,T,test_types)
{
  using dataframe_type = dataframe::dataframe<>;
  dataframe_type dataframe;
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_read_write,T,test_types)
{
  using dataframe_type = dataframe::dataframe<>;
  using ifstream_type = boost::filesystem::ifstream;
  using ofstream_type = boost::filesystem::ofstream;

  using vector_type = std::vector<std::string>;

  using boost::equal;
  using algorithm::unique;
  using dataframe::comment;
  using dataframe::delimiter;
  using dataframe::make_row_proxy;
  using dataframe::skip;
  using dataframe::skip_blank_lines;
  using dataframe::trim_fields;
  using dataframe::trim_whitespace;
  
  ifstream_type is("test.dat");
  ofstream_type os("test.out");

  dataframe_type dataframe;
  is >> comment('#')
     >> skip(2)
     >> skip_blank_lines()
     >> trim_fields()
     >> trim_whitespace()
     >> dataframe;

  vector_type headers = { "name", "description", "value", "sometimes missing", "???" };
  BOOST_CHECK_EQUAL(dataframe.headers().size(),headers.size());  
  BOOST_CHECK(equal(dataframe.headers(),headers));  

  BOOST_CHECK(!dataframe.empty());
  BOOST_CHECK_EQUAL(dataframe.size(),4);
  BOOST_CHECK_EQUAL(dataframe.fields(),5);

  BOOST_CHECK_EQUAL(dataframe.header(0),"name");
  BOOST_CHECK_EQUAL(dataframe.header(1),"description");
  BOOST_CHECK_EQUAL(dataframe.header(2),"value");
  
  BOOST_CHECK_EQUAL(dataframe.field(2).at(0),"+1");
  BOOST_CHECK_EQUAL(dataframe.field(2).at(2),"3.1459");

  BOOST_CHECK_EQUAL(dataframe.field("value").at(0),"+1");
  BOOST_CHECK_EQUAL(dataframe.field("value").at(2),"3.1459");
  
  BOOST_CHECK_EQUAL(dataframe.field<T>(2).at(0),1);
  BOOST_CHECK_EQUAL(dataframe.field<T>(2).at(2),boost::lexical_cast<T>("3.1459"));

  BOOST_CHECK_EQUAL(dataframe.field<T>("value").at(0),1);
  BOOST_CHECK_EQUAL(dataframe.field<T>("value").at(2),boost::lexical_cast<T>("3.1459"));
  
  BOOST_CHECK_EQUAL(dataframe.element(0,2),"+1");
  BOOST_CHECK_EQUAL(dataframe.element(2,2),"3.1459");

  BOOST_CHECK_EQUAL(dataframe.element(0,"value"),"+1");
  BOOST_CHECK_EQUAL(dataframe.element(2,"value"),"3.1459");

  BOOST_CHECK_EQUAL(dataframe.element<T>(0,2),1);
  BOOST_CHECK_EQUAL(dataframe.element<T>(2,2),boost::lexical_cast<T>("3.1459"));

  BOOST_CHECK_EQUAL(dataframe.element<T>(0,"value"),1);
  BOOST_CHECK_EQUAL(dataframe.element<T>(2,"value"),boost::lexical_cast<T>("3.1459"));

  BOOST_CHECK_EQUAL(dataframe(2).at(0),"+1");
  BOOST_CHECK_EQUAL(dataframe(2).at(2),"3.1459");

  BOOST_CHECK_EQUAL(dataframe("value").at(0),"+1");
  BOOST_CHECK_EQUAL(dataframe("value").at(2),"3.1459");

  BOOST_CHECK_EQUAL(dataframe(0,2),"+1");
  BOOST_CHECK_EQUAL(dataframe(2,2),"3.1459");

  BOOST_CHECK_EQUAL(dataframe(0,"value"),"+1");
  BOOST_CHECK_EQUAL(dataframe(2,"value"),"3.1459");

  BOOST_CHECK_EQUAL(dataframe("value").size(),4);
  BOOST_CHECK_EQUAL(unique(dataframe("value")).size(),3);
  BOOST_CHECK_EQUAL(*unique(dataframe("value")).begin(),"+1");
  BOOST_CHECK_EQUAL(*--unique(dataframe("value")).end(),"3.1459");

  BOOST_CHECK_EQUAL(unique(dataframe.field<T>("value")).size(),3);
  BOOST_CHECK_EQUAL(*unique(dataframe.field<T>("value")).begin(),-2);
  BOOST_CHECK_EQUAL(*--unique(dataframe.field<T>("value")).end(),boost::lexical_cast<T>("3.1459"));

  auto row_0 = make_row_proxy(dataframe);
  BOOST_CHECK_EQUAL(row_0(0),"a");
  BOOST_CHECK_EQUAL(row_0(1),"first (*)");
  BOOST_CHECK_EQUAL(row_0("name"),"a");
  BOOST_CHECK_EQUAL(row_0("description"),"first (*)");

  auto row = make_row_proxy(dataframe,2);
  BOOST_CHECK_EQUAL(row(0),"c");
  BOOST_CHECK_EQUAL(row(1),"third");
  BOOST_CHECK_EQUAL(row("name"),"c");
  BOOST_CHECK_EQUAL(row("description"),"third");

  dataframe.insert("new field",{ "1","2","3","4" });
  dataframe.insert("another new field",{ 1,2,3,4 });
  dataframe.insert("a constant field",4,"value");
  dataframe.insert("a varying field",std::vector<std::string>({ "a","b","c","d" }));
  dataframe.insert_row(std::vector<std::string>({"a","b","c","d","e","f","g","h","i"}));
  dataframe.erase_row(3);
  dataframe.erase("???");
  os << delimiter('|') << dataframe;
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_default_factor,T,test_types)
{
  using dataframe_type = dataframe::dataframe<>;
  using factor_type = dataframe::factor<>;

  dataframe_type dataframe;

  dataframe.insert("h1",{ "3.14159", "1", "1", "2", "2", "0" });
  dataframe.factor("h1");
  auto factor = dataframe.element<factor_type>(1,"h1");
  auto other = dataframe.element<factor_type>(0,"h1");

  BOOST_CHECK_EQUAL(factor.size(),4);
  BOOST_CHECK_EQUAL(*(factor.begin()+0),"0");
  BOOST_CHECK_EQUAL(*(factor.begin()+1),"1");
  BOOST_CHECK_EQUAL(*(factor.begin()+2),"2");
  BOOST_CHECK_EQUAL(*(factor.begin()+3),"3.14159");
  BOOST_CHECK(factor.begin()+4 == factor.end());
  BOOST_CHECK_EQUAL(factor.value(),"1");

  BOOST_CHECK(factor.begin() == other.begin());
  BOOST_CHECK(factor.end() == other.end());
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_factor,T,test_types)
{
  using dataframe_type = dataframe::dataframe<>;
  using factor_type = dataframe::factor<T>;

  dataframe_type dataframe;

  dataframe.insert("h1",{ "3.14159", "1", "1", "2", "2", "0" });
  dataframe.factor<T>("h1");
  auto factor = dataframe.element<factor_type>(1,"h1");
  auto other = dataframe.element<factor_type>(0,"h1");

  BOOST_CHECK_EQUAL(factor.size(),4);
  BOOST_CHECK_EQUAL(*(factor.begin()+0),0);
  BOOST_CHECK_EQUAL(*(factor.begin()+1),1);
  BOOST_CHECK_EQUAL(*(factor.begin()+2),2);
  BOOST_CHECK_EQUAL(*(factor.begin()+3),boost::lexical_cast<T>("3.14159"));
  BOOST_CHECK(factor.begin()+4 == factor.end());
  BOOST_CHECK_EQUAL(factor.value(),1);

  BOOST_CHECK(factor.begin() == other.begin());
  BOOST_CHECK(factor.end() == other.end());
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_merge,T,test_types)
{
  using dataframe_type = dataframe::dataframe<>;
  using factor_type = dataframe::factor<T>;

  dataframe_type dataframe;

  dataframe.insert("h1",{ "3.14159", "1", "1", "2", "2", "0" });
  dataframe.insert("h2",{ "1.414", "1", "8", "42", "13", "42" });
  merge(dataframe.factor<T>("h1"),dataframe.factor<T>("h2"));
  auto factor = dataframe.element<factor_type>(1,"h1");
  auto other = dataframe.element<factor_type>(0,"h2");

  BOOST_CHECK_EQUAL(factor.size(),8);
  BOOST_CHECK_EQUAL(*(factor.begin()+0),0);
  BOOST_CHECK_EQUAL(*(factor.begin()+1),1);
  BOOST_CHECK_EQUAL(*(factor.begin()+2),boost::lexical_cast<T>("1.414"));
  BOOST_CHECK_EQUAL(*(factor.begin()+3),boost::lexical_cast<T>("2"));
  BOOST_CHECK_EQUAL(*(factor.begin()+4),boost::lexical_cast<T>("3.14159"));
  BOOST_CHECK(factor.begin()+8 == factor.end());
  BOOST_CHECK_EQUAL(factor.value(),1);
  BOOST_CHECK_EQUAL(other.value(),boost::lexical_cast<T>("1.414"));

  BOOST_CHECK(factor.begin() == other.begin());
  BOOST_CHECK(factor.end() == other.end());
}

