/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#define BOOST_TEST_MODULE graph_model_examples_logistic_regression

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;

#include <asynchronous/signals/type/default_signal_executor_is_sequential.hpp>

#include <iostream>
#include <map>

#include <boost/filesystem/fstream.hpp>
#include <boost/log/sources/severity_logger.hpp>
#include <boost/log/trivial.hpp>

#include <dataframe/dataframe.hpp>
#include <dataframe/element.hpp>
#include <dataframe/read_dataframe.hpp>

#include <graph_model/MCMC/mcmc_signal_policy.hpp>
#include <graph_model/MCMC/metropolis_sampler.hpp>
#include <graph_model/MCMC/sampler_collection.hpp>
#include <graph_model/distribution/bernoulli/bernoulli_distribution.hpp>
#include <graph_model/distribution/normal/normal_distribution.hpp>
#include <graph_model/signals/tags.hpp>
#include <graph_model/vertex/exp.hpp>
#include <graph_model/vertex/expression.hpp>
#include <graph_model/vertex/factor.hpp>
#include <graph_model/vertex/lit.hpp>
#include <graph_model/vertex/markov_blanket.hpp>
#include <graph_model/vertex/sample.hpp>
#include <graph_model/vertex/summarized_random_variable.hpp>

#include <identifiers/x.hpp>
#include <identifiers/y.hpp>
#include <mcmc/mcmc_tally.hpp>

#include <logging/initialize.hpp>

#include <boost/mpl/pair.hpp>
#include <boost/mpl/vector.hpp>

using test_types = boost::mpl::vector<
  boost::mpl::pair<double,long double>
  >;

BOOST_AUTO_TEST_CASE_TEMPLATE(test_graph_model,Pair,test_types)
{
  using variate_type = typename Pair::first;
  using probability_type = typename Pair::second;

  using ifstream_type = boost::filesystem::ifstream;
  using ofstream_type = boost::filesystem::ofstream;

  using factor_type = graph_model::vertex::factor<probability_type>;
  using factor_map_type = std::map<int,factor_type>;

  using expression_type = graph_model::vertex::expression<variate_type>;
  using data_map_type = std::map<int,expression_type>;

  using iteration = graph_model::signals::tag::iteration;
  using accept = graph_model::signals::tag::accept;
  using reject = graph_model::signals::tag::reject;

  using severity_level = boost::log::trivial::severity_level;
  using logger_type = boost::log::sources::severity_logger<severity_level>;

  using graph_model::vertex::lit;

  using dataframe::make_dataframe;
  using dataframe::make_element;

  using graph_model::MCMC::make_mcmc_signal_policy;
  using graph_model::MCMC::metropolis_sampler;
  using graph_model::MCMC::make_sampler_collection;

  using graph_model::accumulators::make_accumulator_set;
  using graph_model::distribution::bernoulli::bernoulli_distribution;
  using graph_model::distribution::normal::normal_distribution;
  using graph_model::vertex::make_summarized_random_variable;

  using graph_model::distribution::parameter::probability;
  using graph_model::distribution::parameter::variate;

  using identifiers::x;
  using identifiers::y;
  using mcmc::make_mcmc_tally;

  namespace accumulators = boost::accumulators;

  logging::initialize();
  logger_type debug_log;

  // construct the random variables to estimate
  auto b0 = make_summarized_random_variable<variate_type, probability_type>();
  b0.reset_identifier("b0").reset_variate(0);
  auto b1 = make_summarized_random_variable<variate_type, probability_type>();
  b1.reset_identifier("b1").reset_variate(0);

  // construct a dataframe of observations
  ifstream_type is("samples.dat");
  auto dataframe = make_dataframe();
  is >> dataframe;

  // construct random variables for the observations and corresponding factors
  data_map_type outcomes;
  data_map_type covariates;
  factor_map_type bernoulli_factors;

  for (auto row : dataframe)
  {
    covariates[row]
      .reset_identifier(x(row))
      .reset_strategy(make_element<variate_type>(dataframe, row, "x"))
      .add_vertex(bernoulli_factors[row]);

    outcomes[row]
      .reset_identifier(y(row))
      .reset_strategy(make_element<variate_type>(dataframe, row, "observed"))
      .add_vertex(bernoulli_factors[row]);

    bernoulli_factors[row]
      .reset_identifier("bernoulli_factors ", row)
      .reset_strategy
      (bernoulli_distribution
       (variate = outcomes[row],
	probability = 1 / ( lit(1) + exp(-(b0 + b1 * lit(covariates[row]))) )))
      .markov_blanket(b0,b1);
  }

  // construct the MCMC components
  auto accumulator = make_accumulator_set<variate_type>();
  auto b0_tally = make_mcmc_tally();
  auto b1_tally = make_mcmc_tally();
  auto samplers = make_sampler_collection();

  auto b0_sampler = metropolis_sampler
    (
     markov_blanket(b0)
     , make_mcmc_signal_policy()
     .template connect_signal<iteration>
     (
      [] {
        static std::size_t i = 0;
        if (i == 0)       std::cout << "iteration: ";
        if (i % 100 == 0) std::cout << " " << i << std::flush; ++i;
      }
      )
     .template connect_signal<accept>(b0_tally.make_acceptor())
     .template connect_signal<reject>(b0_tally.make_rejector())
     )
    .add_rv(b0, lit(b0) + sample(normal_distribution(0,1)));

  auto b1_sampler = metropolis_sampler
    (
     markov_blanket(b1)
     , make_mcmc_signal_policy()
     .template connect_signal<accept>(b1_tally.make_acceptor())
     .template connect_signal<reject>(b1_tally.make_rejector())
     )
    .add_rv(b1, lit(b1) + sample(normal_distribution(0,0.2)));

  // construct an output stream for the MCMC samples and generate samples
  ofstream_type ofstream("mcmc.dat");
  ofstream << "i\tvar\tMCMC_sample\tmean" << std::endl;
  ofstream << 0 << "\tb0\t" << b0.variate() << "\t" << accumulators::mean(b0.accumulator()) << std::endl;
  ofstream << 0 << "\tb1\t" << b1.variate() << "\t" << accumulators::mean(b1.accumulator()) << std::endl;

  GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "MCMC samples:";
  for (int i = 0; i < MCMC_SAMPLE_SIZE; ++i)
    {
      b0_sampler.update();
      b0_sampler.commit();
      ofstream << i+1 << "\tb0\t" << b0.variate() << "\t" << accumulators::mean(b0.accumulator()) << std::endl;

      b1_sampler.update();
      b1_sampler.commit();
      ofstream << i+1 << "\tb1\t" << b1.variate() << "\t" << accumulators::mean(b1.accumulator()) << std::endl;
    }

  // report summary statistics
  std::cout << "\nestimated b0:   " << accumulators::mean(b0.accumulator()) << std::endl;
  std::cout << "estimated b1:   " << accumulators::mean(b1.accumulator()) << std::endl;
  std::cout << "b0 tally: " << b0_tally << std::endl;
  std::cout << "b1 tally: " << b1_tally << std::endl;
  std::cout << std::endl;
}
