/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#define BOOST_TEST_MODULE graph_model_signals_signal

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;

#include <graph_model/signals/signal.hpp>
#include <graph_model/utility/unique_function_collection.hpp>
#include <functional>
#include <vector>
#include <boost/operators.hpp>

struct copyable
  : boost::totally_ordered<copyable>
{
  copyable (int i) : i_(i) {}
  copyable (copyable const&) = default;
  copyable& operator = (copyable const&) = default;

  int operator () () const { return i_; }

  friend auto operator == (copyable const& left, copyable const& right)
  { return left.i_ == right.i_; }
  friend auto operator < (copyable const& left, copyable const& right)
  { return left.i_ < right.i_; }

  friend std::ostream& operator << (std::ostream& os, copyable const& m)
  { return os << m.i_; }

private:
  int i_;
};

struct sum_combiner
{
  auto result () const { return sum_; }

  void operator () () { ++sum_; }
  void operator () (int i) { sum_ += i; }

private:
  int sum_ = 0;
};

template < typename Signature >
struct function_vector
  : std::vector<std::function<Signature>>
{};

BOOST_AUTO_TEST_CASE(test_default_constructor)
{
  using signal_type = graph_model::signals::signal<int()>;
  static_assert(std::is_same<decltype(std::declval<signal_type>()()),int>::value,"");

  signal_type signal;
}

BOOST_AUTO_TEST_CASE(test_insert_lvalue)
{
  using signal_type = graph_model::signals::signal<int()>;
  static_assert(std::is_same<decltype(std::declval<signal_type>()()),int>::value,"");

  copyable one(1);
  copyable two(2);
  
  signal_type signal;
  signal.connect(two);
  signal.connect(one);
}

BOOST_AUTO_TEST_CASE(test_insert_rvalue)
{
  using signal_type = graph_model::signals::signal<int()>;
  static_assert(std::is_same<decltype(std::declval<signal_type>()()),int>::value,"");

  signal_type signal;
  signal.connect(copyable(1));
  signal.connect(copyable(2));
  signal.connect(copyable(1));
  signal.connect(copyable(1));
}

BOOST_AUTO_TEST_CASE(test_eval)
{
  using signal_type = graph_model::signals::signal<int()>;
  static_assert(std::is_same<decltype(std::declval<signal_type>()()),int>::value,"");

  signal_type signal;
  signal.connect(copyable(1));
  signal.connect(copyable(2));
  signal.connect(copyable(1));
  signal.connect(copyable(1));
  signal();
}

BOOST_AUTO_TEST_CASE(test_eval_void_combiner)
{
  using signal_type = graph_model::signals::signal<void(),sum_combiner>;
  static_assert(std::is_same<decltype(std::declval<signal_type>()()),sum_combiner>::value,"");

  signal_type signal;
  signal.connect(copyable(1));
  signal.connect(copyable(2));
  signal.connect(copyable(1));
  signal.connect(copyable(4));
  signal.connect(copyable(4));
  BOOST_CHECK_EQUAL(signal().result(),5);
}

BOOST_AUTO_TEST_CASE(test_eval_void_combiner_unique)
{
  using signal_type
    = graph_model::signals::signal
    <void(),sum_combiner,use_default<>,graph_model::utility::unique_function_collection>;
  static_assert(std::is_same<decltype(std::declval<signal_type>()()),sum_combiner>::value,"");

  signal_type signal;
  signal.connect(copyable(1));
  signal.connect(copyable(2));
  signal.connect(copyable(1));
  signal.connect(copyable(4));
  signal.connect(copyable(4));
  BOOST_CHECK_EQUAL(signal().result(),3);
}

BOOST_AUTO_TEST_CASE(test_eval_combiner)
{
  using signal_type = graph_model::signals::signal<int(),sum_combiner>;
  static_assert(std::is_same<decltype(std::declval<signal_type>()()),sum_combiner>::value,"");

  signal_type signal;
  signal.connect(copyable(1));
  signal.connect(copyable(2));
  signal.connect(copyable(1));
  signal.connect(copyable(4));
  signal.connect(copyable(4));
  BOOST_CHECK_EQUAL(signal().result(),12);
}

BOOST_AUTO_TEST_CASE(test_eval_combiner_unique)
{
  using signal_type
    = graph_model::signals::signal
    <void(),sum_combiner,use_default<>,graph_model::utility::unique_function_collection>;
  static_assert(std::is_same<decltype(std::declval<signal_type>()()),sum_combiner>::value,"");

  signal_type signal;
  signal.connect(copyable(1));
  signal.connect(copyable(2));
  signal.connect(copyable(1));
  signal.connect(copyable(4));
  signal.connect(copyable(4));
  BOOST_CHECK_EQUAL(signal().result(),3);
}

BOOST_AUTO_TEST_CASE(test_function_vector)
{
  using signal_type = graph_model::signals::signal<void(),use_default<>,use_default<>,function_vector>;
  static_assert(std::is_same<decltype(std::declval<signal_type>()()),void>::value,"");

  signal_type signal;
  signal.connect(copyable(1));
  signal.connect(copyable(2));
  signal.connect(copyable(1));
  signal.connect(copyable(1));
  signal();
}

